function  [u, planinfo, plandata] = opss(xroot, model, cfg)
% OPSS Optimistic planning for MDPs / sparsely stochastic systems (OP-MDP, OPSS)
%   U = OPSS(XROOT, MODEL, CFG)
%   [U, PLANINFO] = OPSS(XROOT, MODEL, CFG)
%   [U, PLANINFO, PLANDATA] = OPSS(XROOT, MODEL, CFG)
% Note U may be a single action or a string 'tree' which means more actions should be applied in
% closed loop, following the policy tree returned in plandata. This switch is made based on the
% config field actiontype, see below for commented defaults.
% Version using plain arrays for speed
% Implements tree and memory reuse

% Author: Lucian Busoniu
% Versions:

CFG.gamma = [];                 % discount factor
CFG.n = [];                     % expansion budget
CFG.criterion = 'contrib';      % expansion criterion: based on the whole tree: 'unif'
                                % based on optimistic subtree: 'contrib'; for other modes see
                                % opss_compatibility
CFG.actsel = 'util';            % action selection at root: 'util'; for other modes see opss_compatibility
CFG.L = [];                     % maximum number of successors for each state-action pair
CFG.U = [];                     % discrete, flat action space
CFG.reusetree = 0;              % try reusing previous tree (only when k is set and > 1)
CFG.reusetree_maxsize = 0;      % do not reuse tree if larger (has more nodes) than this; 0 = unlimited size
CFG.usemem = 0;                 % use memory of b-values
CFG.lipschitz = [];             % Lipschitz constant to use in bound estimates
CFG.mem_maxsize = 0;            % 0: no limit, as large as needed, 
                                % <= 10: given as a multiple of the tree size (ignoring root)
                                % > 10: given as number of points in memory
CFG.mem_detailedstats = 0;      % keep more detailed memory stats (hits, improvements, etc.)
CFG.actiontype = 'one';         % 'one' or 'seq' (which in OPSS is a tree); note no support for subsequences
CFG.k = -1;                     % discrete time step
% implementation options
CFG.compat = 0;                 % 0: use default, 1: use structures, 2: use objects
% display config
CFG.verb = 0;                   % verbosity: the higher, the more detailed the messages displayed


persistent Tparent Tchildren Tdepth Tleaf Tx Tf Tr Tterm Tb Tutil TP treesize
persistent Mx Mb memsize
cfg = checkparams(cfg, CFG);

% forward call to the compatibility version if configured to do so, or we are called with a mode
% that is no longer supported in this version
if cfg.compat == 1 || ~any(strcmp(cfg.criterion, {'contrib', 'unif'})) || ~strcmp(cfg.actsel, 'util'),
    [u, planinfo, plandata] = opss_structures(xroot, model, cfg);
    return;
elseif cfg.compat == 2, 
    [u, planinfo, plandata] = opss_objects(xroot, model, cfg);
    return;
end;

if cfg.reusetree && cfg.usemem, error('Simultaneous memory and tree reuse not implemented.'); end;

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.gamma;
structtovars(cfg, {'n', 'criterion', 'L', 'U', 'usemem', 'lipschitz'});
M = size(U, 2);      % number of discrete actions
dimx = model.p;      % number of state vars

% start counting time
t = cputime;

% 1a) Memory initialization: if uninitialized, or we are just starting a new trajectory
% must be done before tree, as otherwise tree may attempt to use existing memory
if usemem && (isempty(memsize) || cfg.k <= 1),
    % always initialized to empty
    Mx = NaN(dimx, 0, 'double');         % state
    Mb = NaN(M, 0, 'double');            % b-values
    memsize = 0;
end;
% reset memory stats
if usemem && cfg.mem_detailedstats, mem_exacthits = 0; mem_improvements = 0; mem_count = 0; end;

% 1b) Tree initialization
if cfg.reusetree,
    % tree must be emptied if:
    reusetree_clear = isempty(treesize) || ...            % no tree has ever been created
        (cfg.reusetree_maxsize > 0 && treesize >= cfg.reusetree_maxsize) || ... % or old tree too large
        cfg.k <= 1;                                 % or we are just starting a trajectory or working offline
end;
if ~cfg.reusetree || reusetree_clear,
    % bound on tree size, accounting for tree built at given iteration, plus maximum reused tree
    % note reusetree_maxsize=0 means "no max size", so we just take the bound w/o tree reuse
    if cfg.reusetree && cfg.reusetree_maxsize > 0,
        maxtreesize = cfg.reusetree_maxsize + M*L*n + 1;
    else
        maxtreesize = M*L*n + 1;
    end;
    % prepare tree data (note 0 = null pointer)
    Tparent = zeros(2, maxtreesize, 'uint32');        % parent and parent action
    Tchildren = zeros(M, L, maxtreesize, 'uint32');   % indices of children for each action
    Tdepth = zeros(1, maxtreesize, 'uint16');         % depth (uint16 as depth > 65536 considered unfeasible)
    Tleaf = false(1, maxtreesize);                    % leaf flag
    Tx = NaN(dimx, maxtreesize, 'double');            % state
    Tf = NaN(1, maxtreesize,'double');                % single-step probability
    Tr = NaN(1, maxtreesize, 'double');               % reward
    Tterm = false(1, maxtreesize);                    % terminal flag
    Tb = NaN(M, maxtreesize, 'double');               % b-values
    Tutil = NaN(M, maxtreesize, 'double');            % u-values
    TP = NaN(1, maxtreesize, 'double');               % probability of reaching this node
    
    % initialize the root node
    root = 1;
    Tparent(:, root) = 0;
    Tchildren(:, :, root) = 0;
    Tdepth(root) = 0;
    Tleaf(root) = true;
    Tx(:, root) = xroot;
    Tf(root) = 1;
    Tr(root) = 0;
    Tterm(root) = false;    % presumably, we are not asked to make a decision in a terminal state
    if usemem && ~isempty(memsize) && memsize > 0,
        if cfg.mem_detailedstats,
            [Tb(:, root), exacthit, improvement] = estimateb(xroot, Mx, Mb, memsize, discount, lipschitz, M, dimx);
            mem_exacthits = mem_exacthits + exacthit;
            mem_improvements = mem_improvements + improvement;
            mem_count = mem_count + 1;
        else
            Tb(:, root) = estimateb(xroot, Mx, Mb, memsize, discount, lipschitz, M, dimx);
        end;
    else
        Tb(:, root) = repmat(1/(1-discount), M, 1);
    end;
    Tutil(:, root) = zeros(M, 1);
    TP(root) = 1;
    
    treesize = 1;
    if cfg.reusetree, subtreesize = 0; end;
else        % reuse tree
    % find nodes containing xroot, choose a min-depth one (large tree to reuse)
    nodesxroot = findflat(xroot, Tx);
    if isempty(nodesxroot), error('Cannot find xroot on old tree'); end;    % shouldn't happen
    [depthmin idepthmin] = min(Tdepth(nodesxroot));
    subtreeroot = nodesxroot(idepthmin);
    % traverse subtree of this node remembering all nodes
    subtree = subtreeroot; frontier = subtreeroot;
    while ~all(Tleaf(frontier)),            % there are still inner nodes to process
        newfrontier = [];
        for node = frontier(~Tleaf(frontier)),     % loop through each inner node
            children = reshape(Tchildren(:, :, node), 1, []);
            newfrontier = [newfrontier children(children > 0)];
        end;
        % the found nodes are added to the subtree, and become the new frontier
        subtree = [subtree newfrontier];
        frontier = newfrontier;
    end;
    subtreesize = length(subtree);
    
    % create the new tree
    % update node pointers
    newpointers = zeros(1, treesize+1, 'uint32'); % the extra zero is special, will be used below
    newpointers(subtree) = 1:subtreesize;
    % note root pointers to parent are 0
    Tparent(:, 1:subtreesize) = [0 newpointers(Tparent(1, subtree(2:end))); 0 Tparent(2, subtree(2:end))];
    % make the zero chidren point to newpointers(treesize+1) = 0 (so they again become zero)
    children = Tchildren(:, :, subtree); children = children(:); children(children == 0) = treesize + 1;
    Tchildren(:, :, 1:subtreesize) = reshape(newpointers(children), M, L, subtreesize);
    % depth decreases
    Tdepth(1:subtreesize) = Tdepth(subtree) - Tdepth(subtreeroot);
    Tleaf(1:subtreesize) = Tleaf(subtree);
    Tx(:, 1:subtreesize) = Tx(:, subtree);
    Tf(1:subtreesize) = Tf(subtree);
    Tr(1:subtreesize) = Tr(subtree);
    Tterm(1:subtreesize) = Tterm(subtree);
    Tb(:, 1:subtreesize) = Tb(:, subtree);
    Tutil(:, 1:subtreesize) = Tutil(:, subtree);
    % probability increases so that the root has probability 1
    TP(1:subtreesize) = TP(subtree) ./ TP(subtreeroot);
    
    % final reassignments
    root = newpointers(subtreeroot);
    treesize = subtreesize;
    if root ~= 1, error('New root is not 1!'); end; % should never happen
    if abs(TP(root)-1) > 1e-4, disp('OPSS: Warning! P(x0) error exceeds threshold; numerical problems?'); end;
    
    % these should never be used again if the code is correct, but set them to zero just in case
    Tparent(:, treesize+1:end) = 0; Tchildren(:, :, treesize+1:end) = 0;
    Tdepth(treesize+1:end) = 0; Tleaf(treesize+1:end) = false;
    Tx(:, treesize+1:end) = NaN; Tf(treesize+1:end) = NaN; Tr(treesize+1:end) = NaN;
    Tterm(treesize+1:end) = false;
    Tb(:, treesize+1:end) = NaN; Tutil(:, treesize+1:end) = NaN; TP(treesize+1:end) = NaN;
    if cfg.verb >= 4,
        visualizetree(Tparent, Tchildren, Tdepth, Tleaf, Tx, Tf, Tr, Tterm, Tb, Tutil, TP, U, discount, [], [], []); 
        set(gcf, sprintf('k=%d: Visualizing reused tree', cfg.k));
        pause; 
    end;  
end;

% 2) Main tree construction loop
for tau = 1:n,
    
    % 2a) find the candidate leaves for expansion
    if criterion(1) == 'c', 
        % optimistic leaves (forward traversal)
        frontier = root;  % start at root
        while ~all(Tleaf(frontier)),            % there are still inner nodes to process
            newnodes = [];
            for node = frontier(~Tleaf(frontier)),     % loop through each inner node
                % find a greedy branch, breaking ties deterministically (i.e. choose first greedy branch)
                [bmax, jmax] = max(Tb(:, node));
                % add children on greedy branch, accounting for the possibility of fewer than L children
                Ljmax = find(Tchildren(jmax, :, node) > 0, 1, 'last');
                newnodes = [newnodes Tchildren(jmax, 1:Ljmax, node)];
            end;
            % new set of nodes is old set of LEAF nodes, plus the newly added nodes
            frontier = [frontier(Tleaf(frontier)) newnodes];
        end;
    elseif criterion(1) == 'u',             % uniform expansion -- just get all the leaves
        frontier = find(Tleaf);
    end;
    
    % 2b) select a leaf node to expand
    % eliminate the terminal nodes and continue with the rest
    frontier = frontier(~Tterm(frontier));
    % if ALL are terminal, then nothing left to do
    if isempty(frontier), break; end;    
    if criterion(1) == 'c',
        % largest contribution to the bound (breaking ties determ)
        % note we do not consider the factor 1/(1-gamma) as it appears everywhere
        % this is the variant used in the ADPRL paper
        c = TP(frontier) .* (discount .^ double(Tdepth(frontier)));
        [cmax, imax] = max(c);
        nodetoexpand = frontier(imax);
    elseif criterion(1) == 'u',     
        % uniform: expand a node with the smallest depth (breaking ties determ)
        [depthmin imin] = min(Tdepth(frontier));
        nodetoexpand = frontier(imin);
    end;
    
    % 2c) expand the chosen node
    Tleaf(nodetoexpand) = false;        % no longer a leaf
    % run through the discrete actions 
    for j = 1:M,
        % use model to determine distribution over next states (+ rewards, terminal flags)
        [x, r, term, f] = feval(model.fun, model, Tx(:, nodetoexpand), U(:, j));
        if any(r < 0) || any(r > 1), error('Rewards must be in [0, 1]'); end;
        % note we allow FEWER than L actions here
        if length(f) > L, error('Too many successors (%d, at most %d allowed)', length(f), L); end;
        % run through the possible next states
        for ell = 1:length(f),
            % find index where new node will be, save it on the children table of current node
            child = treesize + 1;
            Tchildren(j, ell, nodetoexpand) = child;
            
            % create new node
            Tparent(:, child) = [nodetoexpand; j];      % parent is expanded node, along jth action
            Tchildren(:, :, child) = 0;                 % init to "null"
            Tdepth(child) = Tdepth(nodetoexpand) + 1;   % one deeper than the parent
            Tleaf(child) = true;                        % this is a leaf
            Tx(:, child) = x(:, ell);
            Tf(child) = f(ell);
            Tr(child) = r(ell);
            Tterm(child) = term(ell);
            if term(ell),   
                Tb(:, child) = zeros(M, 1);             % values of terminal nodes exactly 0
            elseif usemem && memsize > 0,  
                if cfg.mem_detailedstats,
                    [Tb(:, child), exacthit, improvement] = estimateb(x(:, ell), Mx, Mb, memsize, discount, lipschitz, M, dimx);
                    mem_exacthits = mem_exacthits + exacthit;
                    mem_improvements = mem_improvements + improvement;
                    mem_count = mem_count + 1;
                else
                    Tb(:, child) = estimateb(x(:, ell), Mx, Mb, memsize, discount, lipschitz, M, dimx);
                end;
            else
                Tb(:, child) = repmat(1/(1-discount), M, 1);
            end;
            Tutil(:, child) = zeros(M, 1);
            TP(child) = TP(nodetoexpand) * f(ell);      % prob = parent prob * single-step prob
            
            % we just added a new node
            treesize = treesize + 1;
        end;
    end;
    
    % 2d) update b-values affected by expanded node
    % these are the b-values of the expanded node itself for all the actions, 
    % and those on the state-ACTION path from the expanded added node to the root
    node = nodetoexpand; jrange = 1:M;
    while node ~= 0,       % when node is 0, it means the root has already been processed
        for j = jrange,
            % account for the possibility of fewer successor states than the maximum L
            Lj = find(Tchildren(j, :, node) > 0, 1, 'last');
            children = Tchildren(j, 1:Lj, node);
            Tb(j, node) = Tf(children) * ( Tr(children) + discount * max(Tb(:, children)) )';
            Tutil(j, node) = Tf(children) * ( Tr(children) + discount * max(Tutil(:, children)) )';
            % children is a vector, Tb or Tutil(children) is a matrix with b-values of actions on
            % columns, so the max takes the maximum across actions and returns an L-length vector
            % that can be operated on in dimension sync with the rest
        end;
        jrange = Tparent(2, node);
        node = Tparent(1, node);
    end;

    if cfg.verb >= 4,
        visualizetree(Tparent, Tchildren, Tdepth, Tleaf, Tx, Tf, Tr, Tterm, Tb, Tutil, TP, U, discount, frontier, nodetoexpand, []); 
        if tau == n, set(gcf, 'Name', sprintf('k=%d: Visualizing final tree', cfg.k)); end;
        pause;
    end;        
    
end;    % end main tree construction loop

% 3) Update memory, if using it
if usemem,
    % 3a) Aggregate duplicates (this code is based by removeduplicates.m, and customized)
    % start with existing memory and values on the new tree; ignore leaves as they are derived from
    % the existing memory so by definition they cannot provide better bounds
    Mxtemp = [Mx Tx(:, ~Tleaf(1:treesize))]';
    Mbtemp = [Mb Tb(:, ~Tleaf(1:treesize))]';
    memsize_initial= size(Mxtemp, 1);
    [Mxtemp, irearrange] = sortrows(Mxtemp);
    Mbtemp = Mbtemp(irearrange, :);
    % indices where points are different (up to numerical tolerance) -- are starting points for the index ranges
    ind = find([1; any(abs(diff(Mxtemp, 1)) > 1e-10, 2)]);
    ind = [ind; size(Mxtemp, 1)+1];  % for the last interval
    memsize = length(ind) - 1;
    % init new memory
    Mx = zeros(dimx, memsize); Mb = zeros(M, memsize);
    % since hopefully most of the data will be unique, handle it efficiently
    indsingle = (diff(ind) == 1)';
    Mx(:, indsingle) = Mxtemp(ind([indsingle false]), :)';
    Mb(:, indsingle) = Mbtemp(ind([indsingle false]), :)';
    for i = find(~indsingle),
        % for x, take the first value -- should be fine
        Mx(:, i) = Mxtemp(ind(i), :)'; 
        % for each b(j), choose the minimum bound along all points
        Mb(:, i) = min(Mbtemp(ind(i):ind(i+1)-1, :))';
    end;
    memsize_duplicates = memsize_initial - memsize;
    
    
    % 3b) Remove redundant points (under the assumption of L1 norm)
    Ms = zeros(M, memsize);
    for i = 1:memsize,
        bi = Mb(:, i);      % save b(x_i)
        Mb(:, i) = Inf;     % equivalent to disregarding this point when estimating bound
        % score of each point is the difference between the estimated bound /without/ the point in
        % the set, and its own bound;
        % this score is negative when the sample is redundant, i.e. its bound is larger than
        % resulting from neighbors (or than absolute bound 1/1-gamma)
        Ms(:, i) = estimateb(Mx(:, i), Mx, Mb, memsize, discount, lipschitz, M, dimx) - bi;
        Mb(:, i) = bi;      % restore b(x_i)
    end;
    % points that do not contribute to the aggregate bound are redundant
    % so are points larger than the absolute upper bound 1/1-gamma
    keep = any(Ms > 0);     % note this any works over columns, i.e. actions j
    % remember what we are removing if we'll be plotting it later
    if cfg.verb >= 3, Mxremoved = Mx(:, ~keep); Mbremoved = Mb(:, ~keep);  end; 
    % shrink memory
    Mx = Mx(:, keep);
    Mb = Mb(:, keep);
    memsize_redundant = memsize - size(Mx, 2);
    memsize = size(Mx, 2);

    % 3c) Remove other points sorted by their score to shrink to allowed memory size
    if cfg.mem_maxsize > 0 && cfg.mem_maxsize <= 10,   
        % memory given as a fraction/multiple of tree size; exclude root to make nicer numbers
        mem_maxsize = ceil((treesize - 1) * cfg.mem_maxsize);
    else
        mem_maxsize = cfg.mem_maxsize;
    end;
    if mem_maxsize > 0 && memsize > mem_maxsize,
        % the aggregate score is the maximum of scores for each action
        Ms = max(Ms(:, keep));
        [Ms, irearrange] = sort(Ms, 2, 'descend');
        % remember what we are removing if we'll be plotting it later
        if cfg.verb >= 3, 
            Mxremoved = [Mxremoved Mx(:, irearrange(mem_maxsize+1:end))]; 
            Mbremoved = [Mbremoved Mb(:, irearrange(mem_maxsize+1:end))]; 
        end; 
        Mx = Mx(:, irearrange(1:mem_maxsize));
        Mb = Mb(:, irearrange(1:mem_maxsize));
        memsize_pruned = memsize - mem_maxsize;
        memsize = size(Mx, 2);
    else
        memsize_pruned = 0;
    end;
    
    % visualize including the points that we'll erase
    if usemem && cfg.verb >= 3,
        visualizememory(Mx, Mb, cfg, Mxremoved, Mbremoved); 
        set(gcf, 'Name', sprintf('k=%d: Visualizing memory and any cleaned up points', cfg.k));
        pause; 
    end;        
end;

% 4) Find action at the tree root
% (greedy w.r.t. field actsel, which can be 'b' or 'util')
if cfg.actiontype(1) == 'o',    % one action
    [utilmax, jmax] = max(Tutil(:, root));
    u = U(:, jmax);
else
    u = 'tree';     % signal the tree should be used to compute a closed loop policy
end;
% output stats if requested
if nargout >= 2,
    planinfo.n = tau;
    planinfo.depth = max(Tdepth);
    planinfo.trun = cputime - t;
    planinfo.treesize = treesize;
    planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
        tau, planinfo.depth, planinfo.treesize); 
    if cfg.reusetree,
        planinfo.reusetree_clear = reusetree_clear;
        planinfo.reusetree_size = subtreesize;
        planinfo.message = [planinfo.message sprintf('; reusedtreesize=%d', planinfo.reusetree_size)];
    end;
    if usemem,
        planinfo = copyfields(...
            varstostruct('memsize', 'memsize_initial', 'memsize_duplicates', 'memsize_redundant', 'memsize_pruned'), ...
            planinfo);
        planinfo.message = [planinfo.message sprintf('; memsizes:init=%d/dup=%d/redund=%d/pruned=%d/fin=%d', ...
            memsize_initial, memsize_duplicates, memsize_redundant, memsize_pruned, memsize)];
        if cfg.mem_detailedstats,
            planinfo.mem_exacthits = mem_exacthits;
            planinfo.mem_avgimprovement = mem_improvements / mem_count;
            planinfo.message = [planinfo.message sprintf('; memhits=%d, memavgimpr=%.2f', ...
                mem_exacthits, planinfo.mem_avgimprovement)];
        end;
    end;
end;
% output only actually used tree data (no memory)
if nargout >= 3,
    plandata.Tparent = Tparent(:, 1:treesize);
    plandata.Tchildren = Tchildren(:, :, 1:treesize);
    plandata.Tdepth = Tdepth(1:treesize);
    plandata.Tleaf = Tleaf(1:treesize);
    plandata.Tx = Tx(:, 1:treesize);
    plandata.Tf = Tf(1:treesize);
    plandata.Tr = Tr(1:treesize);
    plandata.Tterm = Tterm(1:treesize);
    plandata.Tb = Tb(:, 1:treesize);
    plandata.Tutil = Tutil(:, 1:treesize);
    plandata.TP = TP(1:treesize);
end;

end % main function


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CORE LOCAL FUNCTIONS

% Estimate b-values
function [b, exacthit, improvement] = estimateb(x, Mx, Mb, memsize, discount, lipschitz, M, dimx)    
    % compute L1 distance to all points in memory, dimension-wise
    dist = zeros(1, memsize);
    for ii = 1:dimx, dist = dist + abs(x(ii) - Mx(ii, 1:memsize)); end;
    dist = lipschitz .* dist;
    % find b-value estimates given by each of these points
    b = zeros(M, memsize);
    for jj = 1:M, b(jj, :) = Mb(jj, 1:memsize) + dist; end;
    % choose the minimum (and a default 1/1-gamma if the minimum is larger)
    b = min(min(b, [], 2), 1/(1-discount));
    % output stats if required
    if nargout > 1,
        exacthit = any(dist == 0);
        improvement = sum(1/(1-discount) - b);
    end;
end     % estimateb


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Auxiliary local functions

% Plot a graphical representation of the current tree
% (argument list is a mess but whatever)
function visualizetree(Tparent, Tchildren, Tdepth, Tleaf, Tx, Tf, Tr, Tterm, Tb, Tutil, TP, ...
    U, discount, ...
    frontier, expandednode, grfront)
persistent figh;
figsize = [800, 600];
if isempty(figh), figh = figurex(struct('size', figsize)); else figure(figh); end;
clf;

c(Tleaf) = TP(Tleaf) .* (discount .^ double(Tdepth(Tleaf))) ./ (1-discount);

% initialize helper variables for nested functions
M = size(U, 2);
nodestoexpand = {};
% build and customize tree
guitree = uitree('Root', rendernode(1));
guitree.Position(3) = figsize(1);   % expand to figure width
% expand nodes
for i = 1:length(nodestoexpand), guitree.expand(nodestoexpand{i}); end;
% done

    % %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % Nested function to render a node recursively
    function guinode = rendernode(i, jp)
    % create label
    if i == 1,  label = 'root: ';
    else        label = ['u=' num2strx(U(:, jp)', 'v') ' (p=' num2str(Tf(i)) ', r=' num2str(Tr(i)) ')  -->  '];
    end;
    % label = [label 'x=' num2strx(Tx(:, i)', 'v')];
    if Tterm(i), label = [label 'T']; end;
    label = [label ' b=' num2strx(Tb(:, i), 'vector') '; P=' num2str(TP(i))];
    if Tleaf(i), label = [label '; c=' num2str(c(i))]; end;
    if i == expandednode, label = ['<i>** ' label ' **</i>']; end;
    if any(i == frontier), label = ['<font color="blue">' label '</font>']; end;
    if any(i == grfront), label = ['<b>' label '</b>']; end;
    if any('<' == label), label = ['<html>' label '</html>']; end;
    % add tree(i) to tree
    guinode = uitreenode(label, label, [], Tleaf(i));
    if ~Tleaf(i),
        % remember we need to expand this node
        nodestoexpand{end+1} = guinode;
        for jj = 1:M,
            Lj = find(Tchildren(jj, :, i) > 0, 1, 'last');
            for ii = 1:Lj, guinode.add(rendernode(Tchildren(jj, ii, i), jj));  end;
        end;
    end;
    end     % function rendernode
end     % function visualizetree


% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Plot a graphical representation of the current memory
function visualizememory(Mx, Mb, cfg, Mxremoved, Mbremoved)

dimx = size(Mx, 1);
memsize = size(Mx, 2);
discount = cfg.gamma;
lipschitz = cfg.lipschitz;
M = size(cfg.U, 2);

if dimx > 2, return; end;        % can only draw up to 2-D data

persistent figh
figsize = [800, 600];
if isempty(figh), figh = figurex(struct('size', figsize)); else figure(figh); end;
clf;

for i = 1:dimx, X{i} = linspace(min(Mx(i, :)), max(Mx(i, :)), 150); end;
Xflat = flat(X);
B = zeros(M, length(Xflat));
for ii = 1:length(Xflat), B(:, ii) = estimateb(Xflat(: , ii), Mx, Mb, memsize, discount, lipschitz, M, dimx); end;

for j = 1:M,
    subplot(M, 1, j);
    if dimx == 1,
        scatter(Mx, Mb(j, :), 'filled'); hold on;
        plot(X, B(j, :), 'k');
        if nargin >= 4,     % highlight removed points
            scatter(Mxremoved, Mbremoved(j, :), 'x', 'MarkerEdgeColor', 'r', 'LineWidth', 2);
        end;
        set(gca, 'xlim', minmax(Mx));
    elseif dimx == 2
        surf(X{1}, X{2}, reshape(B(j, :), length(X{1}), length(X{2}))', 'EdgeColor', 'none', 'FaceAlpha', 0.5); hold on;
        scatter3(Mx(1, :), Mx(2, :), Mb(j, :), 'filled'); 
        colormap(cool);
        if nargin >= 4,     % highlight removed points
            scatter3(Mxremoved(1, :), Mxremoved(2, :), Mbremoved(j, :), 'x', 'MarkerEdgeColor', 'r', 'LineWidth', 2);
        end;
        set(gca, 'xlim', minmax(Mx(1, :)));
        set(gca, 'ylim', minmax(Mx(2, :)));
    end;
end;

end    % function visualizememory


